@@ -11,6 +11,7 @@
 #include "mutex.h"
 #include "arch/arch.h"
 #include "os/os.h"
+#include "helpers.h"
 
 void fio_mutex_remove(struct fio_mutex *mutex)
 {
@@ -64,15 +65,18 @@ struct fio_mutex *fio_mutex_init(int value)
 		log_err("pthread_mutexattr_init: %s\n", strerror(ret));
 		goto err;
 	}
+#ifdef FIO_HAVE_PSHARED_MUTEX
 	ret = pthread_mutexattr_setpshared(&attr, mflag);
 	if (ret) {
 		log_err("pthread_mutexattr_setpshared: %s\n", strerror(ret));
 		goto err;
 	}
+#endif
 
 	pthread_condattr_init(&cond);
+#ifdef FIO_HAVE_PSHARED_MUTEX
 	pthread_condattr_setpshared(&cond, mflag);
-	pthread_condattr_setclock(&cond, CLOCK_MONOTONIC);
+#endif
 	pthread_cond_init(&mutex->cond, &cond);
 
 	ret = pthread_mutex_init(&mutex->lock, &attr);
@@ -95,7 +99,7 @@ int fio_mutex_down_timeout(struct fio_mutex *mutex, unsigned int seconds)
 	struct timespec t;
 	int ret = 0;
 
-	clock_gettime(CLOCK_MONOTONIC, &t);
+	clock_gettime(CLOCK_REALTIME, &t);
 	t.tv_sec += seconds;
 
 	pthread_mutex_lock(&mutex->lock);
