@@ -91,8 +91,8 @@ static void hexdump(void *buffer, int len)
 	int i;
 
 	for (i = 0; i < len; i++)
-		log_info("%02x", p[i]);
-	log_info("\n");
+		log_err("%02x", p[i]);
+	log_err("\n");
 }
 
 /*
@@ -156,243 +156,273 @@ static void *hdr_priv(struct verify_header *hdr)
 	return priv + sizeof(struct verify_header);
 }
 
+/*
+ * Verify container, pass info to verify handlers and allow them to
+ * pass info back in case of error
+ */
+struct vcont {
+	/*
+	 * Input
+	 */
+	struct io_u *io_u;
+	unsigned int hdr_num;
+
+	/*
+	 * Output, only valid in case of error
+	 */
+	const char *name;
+	void *good_crc;
+	void *bad_crc;
+	unsigned int crc_len;
+};
+
+static void log_verify_failure(struct verify_header *hdr, struct vcont *vc)
+{
+	unsigned long long offset;
+
+	offset = vc->io_u->offset;
+	offset += vc->hdr_num * hdr->len;
+	log_err("%.8s: verify failed at file %s offset %llu, length %u\n",
+			vc->name, vc->io_u->file->file_name, offset, hdr->len);
+
+	if (vc->good_crc && vc->bad_crc) {
+		log_err("       Expected CRC: ");
+		hexdump(vc->good_crc, vc->crc_len);
+		log_err("       Received CRC: ");
+		hexdump(vc->bad_crc, vc->crc_len);
+	}
+}
+
 /*
  * Return data area 'header_num'
  */
-static inline void *io_u_verify_off(struct verify_header *hdr,
-				    struct io_u *io_u, unsigned int header_num)
+static inline void *io_u_verify_off(struct verify_header *hdr, struct vcont *vc)
 {
-	return io_u->buf + header_num * hdr->len + hdr_size(hdr);
+	return vc->io_u->buf + vc->hdr_num * hdr->len + hdr_size(hdr);
 }
 
 static int verify_io_u_meta(struct verify_header *hdr, struct thread_data *td,
-			    struct io_u *io_u, unsigned int header_num)
+			    struct vcont *vc)
 {
 	struct vhdr_meta *vh = hdr_priv(hdr);
+	struct io_u *io_u = vc->io_u;
 
 	dprint(FD_VERIFY, "meta verify io_u %p, len %u\n", io_u, hdr->len);
 
-	if (vh->offset != io_u->offset + header_num * td->o.verify_interval) {
-		log_err("meta: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		return EILSEQ;
-	}
+	if (vh->offset == io_u->offset + vc->hdr_num * td->o.verify_interval)
+		return 0;
 
-	return 0;
+	vc->name = "meta";
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_sha512(struct verify_header *hdr, struct io_u *io_u,
-			      unsigned int header_num)
+static int verify_io_u_sha512(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_sha512 *vh = hdr_priv(hdr);
 	uint8_t sha512[128];
 	struct sha512_ctx sha512_ctx = {
 		.buf = sha512,
 	};
 
-	dprint(FD_VERIFY, "sha512 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "sha512 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	sha512_init(&sha512_ctx);
 	sha512_update(&sha512_ctx, p, hdr->len - hdr_size(hdr));
 
-	if (memcmp(vh->sha512, sha512_ctx.buf, sizeof(sha512))) {
-		log_err("sha512: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->sha512, sizeof(vh->sha512));
-		hexdump(sha512_ctx.buf, sizeof(sha512));
-		return EILSEQ;
-	}
+	if (!memcmp(vh->sha512, sha512_ctx.buf, sizeof(sha512)))
+		return 0;
 
-	return 0;
+	vc->name = "sha512";
+	vc->good_crc = vh->sha512;
+	vc->bad_crc = sha512_ctx.buf;
+	vc->crc_len = sizeof(vh->sha512);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_sha256(struct verify_header *hdr, struct io_u *io_u,
-			      unsigned int header_num)
+static int verify_io_u_sha256(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_sha256 *vh = hdr_priv(hdr);
 	uint8_t sha256[64];
 	struct sha256_ctx sha256_ctx = {
 		.buf = sha256,
 	};
 
-	dprint(FD_VERIFY, "sha256 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "sha256 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	sha256_init(&sha256_ctx);
 	sha256_update(&sha256_ctx, p, hdr->len - hdr_size(hdr));
 
-	if (memcmp(vh->sha256, sha256_ctx.buf, sizeof(sha256))) {
-		log_err("sha256: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->sha256, sizeof(vh->sha256));
-		hexdump(sha256_ctx.buf, sizeof(sha256));
-		return EILSEQ;
-	}
+	if (!memcmp(vh->sha256, sha256_ctx.buf, sizeof(sha256)))
+		return 0;
 
-	return 0;
+	vc->name = "sha256";
+	vc->good_crc = vh->sha256;
+	vc->bad_crc = sha256_ctx.buf;
+	vc->crc_len = sizeof(vh->sha256);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_sha1(struct verify_header *hdr, struct io_u *io_u,
-			    unsigned int header_num)
+static int verify_io_u_sha1(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_sha1 *vh = hdr_priv(hdr);
 	uint32_t sha1[5];
 	struct sha1_ctx sha1_ctx = {
 		.H = sha1,
 	};
 
-	dprint(FD_VERIFY, "sha1 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "sha1 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	sha1_init(&sha1_ctx);
 	sha1_update(&sha1_ctx, p, hdr->len - hdr_size(hdr));
 
-	if (memcmp(vh->sha1, sha1_ctx.H, sizeof(sha1))) {
-		log_err("sha1: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->sha1, sizeof(vh->sha1));
-		hexdump(sha1_ctx.H, sizeof(sha1));
-		return EILSEQ;
-	}
+	if (!memcmp(vh->sha1, sha1_ctx.H, sizeof(sha1)))
+		return 0;
 
-	return 0;
+	vc->name = "sha1";
+	vc->good_crc = vh->sha1;
+	vc->bad_crc = sha1_ctx.H;
+	vc->crc_len = sizeof(vh->sha1);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_crc7(struct verify_header *hdr, struct io_u *io_u,
-			    unsigned int header_num)
+static int verify_io_u_crc7(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_crc7 *vh = hdr_priv(hdr);
 	unsigned char c;
 
-	dprint(FD_VERIFY, "crc7 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "crc7 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	c = crc7(p, hdr->len - hdr_size(hdr));
 
-	if (c != vh->crc7) {
-		log_err("crc7: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc7: wanted %x, got %x\n", vh->crc7, c);
-		return EILSEQ;
-	}
+	if (c == vh->crc7)
+		return 0;
 
-	return 0;
+	vc->name = "crc7";
+	vc->good_crc = &vh->crc7;
+	vc->bad_crc = &c;
+	vc->crc_len = 1;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_crc16(struct verify_header *hdr, struct io_u *io_u,
-			     unsigned int header_num)
+static int verify_io_u_crc16(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_crc16 *vh = hdr_priv(hdr);
 	unsigned short c;
 
-	dprint(FD_VERIFY, "crc16 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "crc16 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	c = crc16(p, hdr->len - hdr_size(hdr));
 
-	if (c != vh->crc16) {
-		log_err("crc16: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc16: wanted %x, got %x\n", vh->crc16, c);
-		return EILSEQ;
-	}
+	if (c == vh->crc16)
+		return 0;
 
-	return 0;
+	vc->name = "crc16";
+	vc->good_crc = &vh->crc16;
+	vc->bad_crc = &c;
+	vc->crc_len = 2;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_crc64(struct verify_header *hdr, struct io_u *io_u,
-			     unsigned int header_num)
+static int verify_io_u_crc64(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_crc64 *vh = hdr_priv(hdr);
 	unsigned long long c;
 
-	dprint(FD_VERIFY, "crc64 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "crc64 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	c = crc64(p, hdr->len - hdr_size(hdr));
 
-	if (c != vh->crc64) {
-		log_err("crc64: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len,
-				hdr->len);
-		log_err("crc64: wanted %llx, got %llx\n",
-					(unsigned long long) vh->crc64, c);
-		return EILSEQ;
-	}
+	if (c == vh->crc64)
+		return 0;
 
-	return 0;
+	vc->name = "crc64";
+	vc->good_crc = &vh->crc64;
+	vc->bad_crc = &c;
+	vc->crc_len = 8;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_crc32(struct verify_header *hdr, struct io_u *io_u,
-			     unsigned int header_num)
+static int verify_io_u_crc32(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_crc32 *vh = hdr_priv(hdr);
 	uint32_t c;
 
-	dprint(FD_VERIFY, "crc32 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "crc32 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	c = crc32(p, hdr->len - hdr_size(hdr));
 
-	if (c != vh->crc32) {
-		log_err("crc32: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc32: wanted %x, got %x\n", vh->crc32, c);
-		return EILSEQ;
-	}
+	if (c == vh->crc32)
+		return 0;
 
-	return 0;
+	vc->name = "crc32";
+	vc->good_crc = &vh->crc32;
+	vc->bad_crc = &c;
+	vc->crc_len = 4;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_crc32c(struct verify_header *hdr, struct io_u *io_u,
-			      unsigned int header_num)
+static int verify_io_u_crc32c(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_crc32 *vh = hdr_priv(hdr);
 	uint32_t c;
 
-	dprint(FD_VERIFY, "crc32c verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "crc32c verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	if (hdr->verify_type == VERIFY_CRC32C_INTEL)
 		c = crc32c_intel(p, hdr->len - hdr_size(hdr));
 	else
 		c = crc32c(p, hdr->len - hdr_size(hdr));
 
-	if (c != vh->crc32) {
-		log_err("crc32c: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		log_err("crc32c: wanted %x, got %x\n", vh->crc32, c);
-		return EILSEQ;
-	}
+	if (c == vh->crc32)
+		return 0;
 
-	return 0;
+	vc->name = "crc32c";
+	vc->good_crc = &vh->crc32;
+	vc->bad_crc = &c;
+	vc->crc_len = 4;
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
-static int verify_io_u_md5(struct verify_header *hdr, struct io_u *io_u,
-			   unsigned int header_num)
+static int verify_io_u_md5(struct verify_header *hdr, struct vcont *vc)
 {
-	void *p = io_u_verify_off(hdr, io_u, header_num);
+	void *p = io_u_verify_off(hdr, vc);
 	struct vhdr_md5 *vh = hdr_priv(hdr);
 	uint32_t hash[MD5_HASH_WORDS];
 	struct md5_ctx md5_ctx = {
 		.hash = hash,
 	};
 
-	dprint(FD_VERIFY, "md5 verify io_u %p, len %u\n", io_u, hdr->len);
+	dprint(FD_VERIFY, "md5 verify io_u %p, len %u\n", vc->io_u, hdr->len);
 
 	md5_init(&md5_ctx);
 	md5_update(&md5_ctx, p, hdr->len - hdr_size(hdr));
 
-	if (memcmp(vh->md5_digest, md5_ctx.hash, sizeof(hash))) {
-		log_err("md5: verify failed at %llu/%u\n",
-				io_u->offset + header_num * hdr->len, hdr->len);
-		hexdump(vh->md5_digest, sizeof(vh->md5_digest));
-		hexdump(md5_ctx.hash, sizeof(hash));
-		return EILSEQ;
-	}
+	if (!memcmp(vh->md5_digest, md5_ctx.hash, sizeof(hash)))
+		return 0;
 
-	return 0;
+	vc->name = "md5";
+	vc->good_crc = vh->md5_digest;
+	vc->bad_crc = md5_ctx.hash;
+	vc->crc_len = sizeof(hash);
+	log_verify_failure(hdr, vc);
+	return EILSEQ;
 }
 
 static unsigned int hweight8(unsigned int w)
@@ -468,19 +498,24 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 	ret = 0;
 	for (p = io_u->buf; p < io_u->buf + io_u->buflen;
 	     p += hdr_inc, hdr_num++) {
-		if (ret && td->o.verify_fatal) {
-			td->terminate = 1;
+		struct vcont vc = {
+			.io_u		= io_u,
+			.hdr_num	= hdr_num,
+		};
+
+		if (ret && td->o.verify_fatal)
 			break;
-		}
+
 		hdr_size = __hdr_size(td->o.verify);
 		if (td->o.verify_offset)
 			memswp(p, p + td->o.verify_offset, hdr_size);
 		hdr = p;
 
 		if (hdr->fio_magic != FIO_HDR_MAGIC) {
-			log_err("Bad verify header %x at %llu\n",
-					hdr->fio_magic,
-					io_u->offset + hdr_num * hdr->len);
+			log_err("verify: bad magic header %x, wanted %x at file %s offset %llu, length %u\n",
+				hdr->fio_magic, FIO_HDR_MAGIC,
+				io_u->file->file_name,
+				io_u->offset + hdr_num * hdr->len, hdr->len);
 			return EILSEQ;
 		}
 
@@ -492,50 +527,53 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 				  p + hdr_size,
 				  hdr_inc - hdr_size,
 				  hdr_size % td->o.verify_pattern_bytes);
+
+			if (ret) {
+				log_err("pattern: verify failed at file %s offset %llu, length %u\n",
+					io_u->file->file_name,
+					io_u->offset + hdr_num * hdr->len,
+					hdr->len);
+			}
+
 			/*
 			 * Also verify the meta data, if applicable
 			 */
 			if (hdr->verify_type == VERIFY_META)
-				ret |= verify_io_u_meta(hdr, td, io_u, hdr_num);
-
-			if (ret)
-				log_err("fio: verify failed at %llu/%u\n",
-					io_u->offset + hdr_num * hdr->len,
-					hdr->len);
+				ret |= verify_io_u_meta(hdr, td, &vc);
 			continue;
 		}
 
 		switch (hdr->verify_type) {
 		case VERIFY_MD5:
-			ret = verify_io_u_md5(hdr, io_u, hdr_num);
+			ret = verify_io_u_md5(hdr, &vc);
 			break;
 		case VERIFY_CRC64:
-			ret = verify_io_u_crc64(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc64(hdr, &vc);
 			break;
 		case VERIFY_CRC32C:
 		case VERIFY_CRC32C_INTEL:
-			ret = verify_io_u_crc32c(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc32c(hdr, &vc);
 			break;
 		case VERIFY_CRC32:
-			ret = verify_io_u_crc32(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc32(hdr, &vc);
 			break;
 		case VERIFY_CRC16:
-			ret = verify_io_u_crc16(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc16(hdr, &vc);
 			break;
 		case VERIFY_CRC7:
-			ret = verify_io_u_crc7(hdr, io_u, hdr_num);
+			ret = verify_io_u_crc7(hdr, &vc);
 			break;
 		case VERIFY_SHA256:
-			ret = verify_io_u_sha256(hdr, io_u, hdr_num);
+			ret = verify_io_u_sha256(hdr, &vc);
 			break;
 		case VERIFY_SHA512:
-			ret = verify_io_u_sha512(hdr, io_u, hdr_num);
+			ret = verify_io_u_sha512(hdr, &vc);
 			break;
 		case VERIFY_META:
-			ret = verify_io_u_meta(hdr, td, io_u, hdr_num);
+			ret = verify_io_u_meta(hdr, td, &vc);
 			break;
 		case VERIFY_SHA1:
-			ret = verify_io_u_sha1(hdr, io_u, hdr_num);
+			ret = verify_io_u_sha1(hdr, &vc);
 			break;
 		default:
 			log_err("Bad verify type %u\n", hdr->verify_type);
@@ -543,6 +581,9 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 		}
 	}
 
+	if (ret && td->o.verify_fatal)
+		td->terminate = 1;
+
 	return ret;
 }
 
@@ -840,7 +881,8 @@ static void *verify_async_thread(void *data)
 
 	if (ret) {
 		td_verror(td, ret, "async_verify");
-		td->terminate = 1;
+		if (td->o.verify_fatal)
+			td->terminate = 1;
 	}
 
 done:
